# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import random

from absl.testing import absltest
import numpy as np

from open_spiel.python.algorithms import async_mcts
from open_spiel.python.algorithms import evaluate_bots
import pyspiel

UCT_C = math.sqrt(2)


def _get_action(state, action_str):
  for action in state.legal_actions():
    if action_str == state.action_to_string(state.current_player(), action):
      return action
  raise ValueError("invalid action string: {}".format(action_str))


def search_tic_tac_toe_state(initial_actions):
  game = pyspiel.load_game("tic_tac_toe")
  state = game.new_initial_state()
  for action_str in initial_actions.split(" "):
    state.apply_action(_get_action(state, action_str))
  rng = np.random.RandomState(42)
  bot = async_mcts.MCTSBot(
      game,
      UCT_C,
      max_simulations=10000,
      solve=True,
      random_state=rng,
      evaluator=async_mcts.RandomRolloutEvaluator(random_state=rng),
  )
  return bot.mcts_search(state), state


def make_node(action, player=0, prior=1, **kwargs):
  node = async_mcts.SearchNode(action, player, prior)
  for k, v in kwargs.items():
    setattr(node, k, v)
  return node


class MctsBotTest(absltest.TestCase):

  def test_can_play_tic_tac_toe(self):
    game = pyspiel.load_game("tic_tac_toe")
    max_simulations = 100
    evaluator = async_mcts.RandomRolloutEvaluator()
    bots = [
        async_mcts.MCTSBot(game, UCT_C, max_simulations, evaluator),
        async_mcts.MCTSBot(game, UCT_C, max_simulations, evaluator),
    ]
    v = evaluate_bots.evaluate_bots(game.new_initial_state(), bots, np.random)
    self.assertEqual(v[0] + v[1], 0)

  def test_can_play_both_sides(self):
    game = pyspiel.load_game("tic_tac_toe")
    bot = async_mcts.MCTSBot(
        game,
        UCT_C,
        max_simulations=100,
        evaluator=async_mcts.RandomRolloutEvaluator(),
    )
    bots = [bot, bot]
    v = evaluate_bots.evaluate_bots(game.new_initial_state(), bots, np.random)
    self.assertEqual(v[0] + v[1], 0)

  def test_can_play_single_player(self):
    game = pyspiel.load_game("catch")
    max_simulations = 100
    evaluator = async_mcts.RandomRolloutEvaluator()
    bots = [async_mcts.MCTSBot(game, UCT_C, max_simulations, evaluator)]
    v = evaluate_bots.evaluate_bots(game.new_initial_state(), bots, np.random)
    self.assertGreater(v[0], 0)

  def test_throws_on_simultaneous_game(self):
    game = pyspiel.load_game("matrix_mp")
    evaluator = async_mcts.RandomRolloutEvaluator()
    with self.assertRaises(ValueError):
      async_mcts.MCTSBot(game, UCT_C, max_simulations=100, evaluator=evaluator)

  def test_can_play_three_player_stochastic_games(self):
    game = pyspiel.load_game("pig(players=3,winscore=20,horizon=30)")
    max_simulations = 100
    evaluator = async_mcts.RandomRolloutEvaluator()
    bots = [
        async_mcts.MCTSBot(game, UCT_C, max_simulations, evaluator),
        async_mcts.MCTSBot(game, UCT_C, max_simulations, evaluator),
        async_mcts.MCTSBot(game, UCT_C, max_simulations, evaluator),
    ]
    v = evaluate_bots.evaluate_bots(game.new_initial_state(), bots, np.random)
    self.assertEqual(sum(v), 0)

  def assertBestChild(self, choice, children):
    # If this causes flakiness, the key in `SearchNode.best_child` is bad.
    random.shuffle(children)
    root = make_node(-1, children=children)
    self.assertEqual(root.best_child().action, choice)

  def test_choose_most_visited_when_not_solved(self):
    self.assertBestChild(
        0,
        [
            make_node(0, explore_count=50, total_reward=30),
            make_node(1, explore_count=40, total_reward=40),
        ],
    )

  def test_choose_win_over_most_visited(self):
    self.assertBestChild(
        1,
        [
            make_node(0, explore_count=50, total_reward=30),
            make_node(1, explore_count=40, total_reward=40, outcome=[1]),
        ],
    )

  def test_choose_best_over_good(self):
    self.assertBestChild(
        1,
        [
            make_node(0, explore_count=50, total_reward=30, outcome=[0.5]),
            make_node(1, explore_count=40, total_reward=40, outcome=[0.8]),
        ],
    )

  def test_choose_bad_over_worst(self):
    self.assertBestChild(
        0,
        [
            make_node(0, explore_count=50, total_reward=30, outcome=[-0.5]),
            make_node(1, explore_count=40, total_reward=40, outcome=[-0.8]),
        ],
    )

  def test_choose_positive_reward_over_promising(self):
    self.assertBestChild(
        1,
        [
            make_node(0, explore_count=50, total_reward=40),  # more promising
            make_node(
                1, explore_count=10, total_reward=1, outcome=[0.1]
            ),  # solved
        ],
    )

  def test_choose_most_visited_over_loss(self):
    self.assertBestChild(
        0,
        [
            make_node(0, explore_count=50, total_reward=30),
            make_node(1, explore_count=40, total_reward=40, outcome=[-1]),
        ],
    )

  def test_choose_most_visited_over_draw(self):
    self.assertBestChild(
        0,
        [
            make_node(0, explore_count=50, total_reward=30),
            make_node(1, explore_count=40, total_reward=40, outcome=[0]),
        ],
    )

  def test_choose_uncertainty_over_most_visited_loss(self):
    self.assertBestChild(
        1,
        [
            make_node(0, explore_count=50, total_reward=30, outcome=[-1]),
            make_node(1, explore_count=40, total_reward=40),
        ],
    )

  def test_choose_slowest_loss(self):
    self.assertBestChild(
        1,
        [
            make_node(0, explore_count=50, total_reward=10, outcome=[-1]),
            make_node(1, explore_count=60, total_reward=15, outcome=[-1]),
        ],
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generate a dummy trajectory and compute the distribution of a policy."""
# pylint: disable=unused-import
from typing import Sequence

from absl import app
from absl import flags
import numpy as np

from open_spiel.python import policy
from open_spiel.python.mfg import games
from open_spiel.python.mfg.algorithms import best_response_value
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import fictitious_play
from open_spiel.python.mfg.algorithms import greedy_policy
from open_spiel.python.mfg.algorithms import mirror_descent
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.algorithms import policy_value
from open_spiel.python.mfg.games import linear_quadratic
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_string('game', 'mean_field_lin_quad', 'Game to use.')
flags.DEFINE_integer('size', 10, 'Number of states.')
flags.DEFINE_integer('horizon', 5, 'Horizon size.')
flags.DEFINE_float('dt', 1.0, 'Delta t.')
flags.DEFINE_integer('n_actions_per_side', 3,
                     'Number actions per side (Total num actions = 2*x+1).')
flags.DEFINE_float('volatility', 1.0, 'Action noise.')
flags.DEFINE_float('learning_rate', 0.01, 'OMD learning rate.')


def get_l1_distribution_dist(mu1, mu2):
  mu1d = mu1.distribution
  mu2d = mu2.distribution
  states = set(list(mu1d.keys()) + list(mu2d.keys()))
  return sum([abs(mu1d.get(a, 0.0) - mu2d.get(a, 0.0)) for a in states
             ]) * FLAGS.dt / FLAGS.horizon


class LinearPolicy(policy.Policy):
  """Project values on the policy simplex."""

  def __init__(self, game, player_ids):  # pylint:disable=useless-super-delegation
    """Initializes the projected policy.

    Args:
      game: The game to analyze.
      player_ids: list of player ids for which this policy applies; each should
        be in the range 0..game.num_players()-1.
    """
    super(LinearPolicy, self).__init__(game, player_ids)

  def action_probabilities(self, state, player_id=None):
    mu_bar_t = state.distribution_average()
    x_t = state.x
    q = state.cross_q
    n_actions_per_side = state.n_actions_per_side
    lin_action = (q + state.eta_t()) * (mu_bar_t - x_t)
    action = n_actions_per_side + min(
        n_actions_per_side, max(round(lin_action), -n_actions_per_side))
    action_prob = [(a, 0.0) for a in state.legal_actions()]
    action_prob[action] = (action, 1.0)
    return dict(action_prob)


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')
  mfg_game = pyspiel.load_game(
      FLAGS.game, {
          'dt': FLAGS.dt,
          'size': FLAGS.size,
          'horizon': FLAGS.horizon,
          'n_actions_per_side': FLAGS.n_actions_per_side,
          'volatility': FLAGS.volatility
      })

  uniform_policy = policy.UniformRandomPolicy(mfg_game)
  nash_conv_fp = nash_conv.NashConv(mfg_game, uniform_policy)
  print('Uniform Policy Nashconv:', nash_conv_fp.nash_conv())

  # Optimal control in the continuous setting.
  theoretical_control = LinearPolicy(mfg_game,
                                     list(range(mfg_game.num_players())))
  theoretical_distribution = distribution.DistributionPolicy(
      mfg_game, theoretical_control)
  discretized_optimal_value = policy_value.PolicyValue(
      mfg_game, theoretical_distribution,
      theoretical_control).eval_state(mfg_game.new_initial_state())

  th_expl = nash_conv.NashConv(mfg_game, theoretical_control).nash_conv()
  print('Theoretical policy NashConv : {}'.format(th_expl))
  print('Theoretical policy Value : {}'.format(discretized_optimal_value))

  fp = fictitious_play.FictitiousPlay(mfg_game)
  md = mirror_descent.MirrorDescent(mfg_game)
  for j in range(1000):
    print('\n\nIteration', j, '\n')
    fp.iteration()
    fp_policy = fp.get_policy()
    nash_conv_fp = nash_conv.NashConv(mfg_game, fp_policy)
    print('Nashconv of the current FP policy', nash_conv_fp.nash_conv())
    fp_current_distribution = distribution.DistributionPolicy(
        mfg_game, fp.get_policy())
    fp_l1_dist = get_l1_distribution_dist(fp_current_distribution,
                                          theoretical_distribution)
    print(
        'L1 distance between FP and theoretical policy : {}'.format(fp_l1_dist))
    md.iteration()
    md_policy = md.get_policy()
    nash_conv_md = nash_conv.NashConv(mfg_game, md_policy)

    print('')

    print('Nashconv of the current MD policy', nash_conv_md.nash_conv())
    md_current_distribution = md._distribution  # pylint:disable=protected-access
    md_l1_dist = get_l1_distribution_dist(md_current_distribution,
                                          theoretical_distribution)
    print('L1 distance between OMD and theoretical policy : {}'.format(
        md_l1_dist))


if __name__ == '__main__':
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Representation of a value for a game.

This is a standard representation for passing value functions into algorithms,
with currently the following implementations:

The main way of using a value is to call `value(state)`
or `value(state, action)`.

We will prevent calling a value on a state action on a MEAN_FIELD state.

The state can be a pyspiel.State object or its string representation. For a
particular ValueFunction instance, you should use only one or the other. The
behavior may be undefined for mixed usage depending on the implementation.
"""

import collections
from typing import Union

import pyspiel

ValueFunctionState = Union[pyspiel.State, str]


class ValueFunction(object):
  """Base class for values.

  A ValueFunction is something that returns a value given
  a state of the world or a state and an action.

  Attributes:
    game: the game for which this ValueFunction derives
  """

  def __init__(self, game):
    """Initializes a value.

    Args:
      game: the game for which this value derives
    """
    self.game = game

  def value(self, state: ValueFunctionState, action=None) -> float:
    """Returns a float representing a value.

    Args:
      state: A `pyspiel.State` object or its string representation.
      action: may be None or a legal action

    Returns:
      A value for the state (and eventuallu state action pair).
    """
    raise NotImplementedError()

  def __call__(self, state: ValueFunctionState, action=None) -> float:
    """Turns the value into a callable.

    Args:
      state: A `pyspiel.State` object or its string representation.
      action: may be None or a legal action

    Returns:
      Float: the value of the state or the state action pair.
    """
    return self.value(state, action=action)

  def set_value(self, state: ValueFunctionState, value: float, action=None):
    """Sets the value of the state.

    Args:
      state: A `pyspiel.State` object or its string representation.
      value: Value of the state.
      action: may be None or a legal action
    """
    raise NotImplementedError()

  def has(self, state: ValueFunctionState, action=None) -> bool:
    """Returns true if state(-action) has an explicit value.

    Args:
      state: A `pyspiel.State` object or its string representation.
      action: may be None or a legal action

    Returns:
      True if there is an explicitly specified value.
    """
    raise NotImplementedError()

  def add_value(self, state, value: float, action=None):
    """Adds the value to the current value of the state.

    Args:
      state: A `pyspiel.State` object or its string representation.
      value: Value to add.
      action: may be None or a legal action
    """
    self.set_value(
        state, self.value(state, action=action) + value, action=action)


class TabularValueFunction(ValueFunction):
  """Tabular value function backed by a dictionary."""

  def __init__(self, game):
    super().__init__(game)
    self._values = collections.defaultdict(float)

  def value(self, state: ValueFunctionState, action=None):
    return self._values[(state, action)]

  def set_value(self, state: ValueFunctionState, value: float, action=None):
    self._values[(state, action)] = value

  def has(self, state: ValueFunctionState, action=None):
    return (state, action) in self._values



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Single Transferrable Vote (STV) method.

Based on https://en.wikipedia.org/wiki/Single_transferable_vote.
"""

from typing import Dict, List, Union
from open_spiel.python.voting import base


class MutableVote(object):
  """A mutable vote annotated with the current preferred alternative.
 
  This is used to keep track of votes and which index (into the preference list)
  is currently active, i.e. the most preferred. When votes get used to determine
  winners or elimintations, some of these votes get "transfered" down to the
  next alternative. To transfer the vote, the index here is incremented to
  indicate that this vote is now representing a vote for the next highest
  alternative.
  """

  def __init__(self, idx: int, weight: int, vote: List[base.AlternativeId]):
    self.idx = idx
    self.weight = weight
    self.vote = vote


class STVVoting(base.AbstractVotingMethod):
  """Implements STV method."""

  def __init__(
      self, num_winners: Union[int, None] = None, verbose: bool = False
  ):
    """Construct an instance of STV with the specified number of winners.

    Args:
      num_winners: number of winners. Should be less than number of
          alternatives (m). If not specified, defaults to int(m/2).
      verbose: whether or not to print debug information as STV is running.
    """
    self._num_winners = num_winners
    self._verbose = verbose

  def name(self) -> str:
    return f"single_transferable_vote(num_winners={self._num_winners})"

  def _is_still_active(
      self,
      alternative: base.AlternativeId,
      winners: List[base.AlternativeId],
      losers: List[base.AlternativeId],
  ) -> bool:
    """Returns whether the alternative is still in the running."""
    return alternative not in winners and alternative not in losers

  def _next_idx_in_the_running(
      self,
      mutable_vote: MutableVote,
      winners: List[base.AlternativeId],
      losers: List[base.AlternativeId],
  ) -> int:
    """"Returns the next index in the list that is still in the running."""
    new_idx = mutable_vote.idx + 1
    while (new_idx < len(mutable_vote.vote) and
           not self._is_still_active(mutable_vote.vote[new_idx], winners,
                                     losers)):
      new_idx += 1
    return new_idx

  def _initial_scores_for_round(
      self,
      profile: base.PreferenceProfile,
      winners: List[base.AlternativeId],
      losers: List[base.AlternativeId],
  ) -> Dict[base.AlternativeId, float]:
    """Returns round's initial scores for alternatives still in the running."""
    alt_scores = {}
    for alt in profile.alternatives:
      if self._is_still_active(alt, winners, losers):
        alt_scores[alt] = 0
    return alt_scores

  def _remove_winning_votes(
      self,
      winning_alt: base.AlternativeId,
      num_to_remove: int,
      all_votes: List[MutableVote],
  ):
    while num_to_remove > 0:
      for mutable_vote in all_votes:
        if (mutable_vote.idx < len(mutable_vote.vote) and
            mutable_vote.vote[mutable_vote.idx] == winning_alt):
          removing_now = min(mutable_vote.weight, num_to_remove)
          mutable_vote.weight -= removing_now
          num_to_remove -= removing_now
        if num_to_remove == 0:
          break

  def run_election(self, profile: base.PreferenceProfile) -> base.RankOutcome:
    assert self.is_valid_profile(profile)
    winners = []
    losers = []
    winner_scores = []
    loser_scores = []
    votes = profile.votes
    total_votes = profile.total_weight()
    m = profile.num_alternatives()
    num_winners = self._num_winners
    if num_winners is None:
      num_winners = int(m/2)
      if self._verbose:
        print("Warning: number of winners not specified." +
              f"Choosing {num_winners}")
    assert num_winners < m
    quota = int(total_votes / float(num_winners + 1) + 1)
    # Table holds a list of the IndexAndWeightedVote. The index corresponds to
    # the current alternative that this vote is representing. They all start at
    # 0 at the start, corresponding to their highest preference, and they get
    # incremented as they become used up.
    all_votes: List[MutableVote] = []
    for vote in votes:
      all_votes.append(MutableVote(idx=0, weight=vote.weight, vote=vote.vote))
    while len(winners) + len(losers) < m:
      scores = self._initial_scores_for_round(profile, winners, losers)
      for mutable_vote in all_votes:
        if (mutable_vote.idx < len(mutable_vote.vote) and
            mutable_vote.weight > 0):
          alt = mutable_vote.vote[mutable_vote.idx]
          scores[alt] += mutable_vote.weight
      sorted_scores = sorted(scores.items(), key=lambda item: item[1],
                             reverse=True)
      best_score = sorted_scores[0][1]
      if best_score >= quota:
        # Quota reached. A candidate wins!
        if self._verbose:
          print(f"Quota {quota} reached. Candidate {sorted_scores[0][0]} wins!")
        winning_alt = sorted_scores[0][0]
        winners.append(winning_alt)
        winner_scores.append(best_score)
        surplus = sorted_scores[0][1] - quota
        # Remove votes that contributed to the winner, up to the quota.
        self._remove_winning_votes(winning_alt, quota, all_votes)
        # Then, convert all the rest.
        num_converted = 0
        for mutable_vote in all_votes:
          if (mutable_vote.idx < len(mutable_vote.vote) and
              mutable_vote.vote[mutable_vote.idx] == winning_alt and
              mutable_vote.weight > 0):
            # find the next one in the list still in the running.
            new_idx = self._next_idx_in_the_running(mutable_vote, winners,
                                                    losers)
            mutable_vote.idx = new_idx
            num_converted += mutable_vote.weight
        assert num_converted == surplus
      else:
        # No winner, eliminate the bottom candidate.
        eliminated_alt = sorted_scores[-1][0]
        eliminated_score = sorted_scores[-1][1]
        if self._verbose:
          print(f"No winner. Quota = {quota}. Eliminating candidate: " +
                f"{eliminated_alt} with score: {eliminated_score}")
        elim_count = sorted_scores[-1][1]
        losers.insert(0, eliminated_alt)
        loser_scores.insert(0, eliminated_score)
        # All of the votes with this alternative as the top is converted.
        votes_counted = 0
        for mutable_vote in all_votes:
          if (mutable_vote.idx < len(mutable_vote.vote) and
              mutable_vote.vote[mutable_vote.idx] == eliminated_alt and
              mutable_vote.weight > 0):
            # find the next one in the list still in the running.
            new_idx = self._next_idx_in_the_running(mutable_vote, winners,
                                                    losers)
            mutable_vote.idx = new_idx
            votes_counted += mutable_vote.weight
        assert votes_counted == elim_count
    ranking = winners + losers
    scores = []
    win_score_base = profile.num_alternatives() * 2
    lose_score_base = profile.num_alternatives()
    for winner_score in winner_scores:
      scores.append(float(str(win_score_base) + "." + str(winner_score)))
      win_score_base -= 1
    for loser_score in loser_scores:
      scores.append(float(str(lose_score_base) + "." + str(loser_score)))
      lose_score_base -= 1
    outcome = base.RankOutcome(rankings=ranking, scores=scores)
    return outcome



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Elo rating system wrapper."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python.voting import base
from open_spiel.python.voting import elo


class EloTest(parameterized.TestCase):

  def test_meeple_pentathlon(self):
    alternatives = ["A", "B", "C"]
    votes = [
        base.WeightedVote(1, ["A", "B", "C"]),
        base.WeightedVote(1, ["A", "C", "B"]),
        base.WeightedVote(2, ["C", "A", "B"]),
        base.WeightedVote(1, ["B", "C", "A"]),
    ]
    profile = base.PreferenceProfile(votes=votes, alternatives=alternatives)
    ratings_map = elo.compute_ratings_from_preference_profile(profile)
    self.assertAlmostEqual(ratings_map["C"], ratings_map["A"])
    self.assertLess(ratings_map["B"], ratings_map["C"])


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Solving strong Stackelberg equilibrium based on linear programming.

Based on [1] "Computing the Optimal Strategy to Commit to", Conitzer & Sandholm,
EC'06
"""

import cvxpy as cp
import numpy as np

from open_spiel.python.algorithms.projected_replicator_dynamics import _simplex_projection
from open_spiel.python.egt.utils import game_payoffs_array


def solve_stackelberg(game, is_first_leader=True):
  """Solves the optimal mixed strategty to commit to for the leader.

  Args:
    game: a pyspiel game,
    is_first_leader: if true, then player 0 is the leader, o.w. player 1 is
      the leader.

  Returns:
    (player0 strategy, player1 strategy, player0 payoff, player1 payoff) at an
    SSE.
  """
  p_mat = game_payoffs_array(game)
  assert len(p_mat) == 2
  if is_first_leader:
    leader_payoff, follower_payoff = p_mat[0], p_mat[1]
  else:
    leader_payoff, follower_payoff = p_mat[1].T, p_mat[0].T

  num_leader_strategies, num_follower_strategies = leader_payoff.shape

  leader_eq_value = -float("inf")
  follower_eq_value = None
  leader_eq_strategy = None
  follower_eq_strategy = None

  for t in range(num_follower_strategies):
    p_s = cp.Variable(num_leader_strategies, nonneg=True)
    constraints = [p_s <= 1, cp.sum(p_s) == 1]
    for t_ in range(num_follower_strategies):
      if t_ == t:
        continue
      constraints.append(
          p_s @ follower_payoff[:, t_] <= p_s @ follower_payoff[:, t]
      )
    prob = cp.Problem(cp.Maximize(p_s @ leader_payoff[:, t]), constraints)
    prob.solve()
    p_s_value = p_s.value
    if p_s_value is None:
      continue
    leader_strategy = _simplex_projection(p_s.value.reshape(-1)).reshape(-1, 1)
    leader_value = leader_strategy.T.dot(leader_payoff)[0, t]
    if leader_value > leader_eq_value:
      leader_eq_strategy = leader_strategy
      follower_eq_strategy = t
      leader_eq_value = leader_value
      follower_eq_value = leader_strategy.T.dot(follower_payoff)[0, t]

  assert leader_eq_strategy is not None, p_mat
  if is_first_leader:
    return (
        leader_eq_strategy.reshape(-1),
        np.identity(num_follower_strategies)[follower_eq_strategy],
        leader_eq_value,
        follower_eq_value,
    )
  else:
    return (np.identity(num_follower_strategies)[follower_eq_strategy],
            leader_eq_strategy.reshape(-1), follower_eq_value, leader_eq_value)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Nash averaging.

Based on https://arxiv.org/abs/1806.02643. An axiomatic strategy evaluation
metric for Agent-vs-Agent or Agent-vs-Task two-player zero-sum games.
"""

import cvxpy as cp
import numpy as np

from open_spiel.python.egt.utils import game_payoffs_array


def _max_entropy_symmetric_nash(p_mat, eps=1e-9):
  """Solves for the maxent symmetric nash for symmetric 2P zero-sum games.

    Using convex programming:
      min p^Tlog(p)
      s.t.
      p_mat.dot(p) <= 0, since game value must be 0
      p >= 0
      1^T * p = 1

  Args:
    p_mat: an N*N anti-symmetric payoff matrix for the row player
    eps: minimum probability threshold

  Returns:
    p*: a maxent symmetric nash
  """
  assert np.array_equal(p_mat, -p_mat.T) and eps >= 0 and eps <= 0.5
  n = len(p_mat)
  x = cp.Variable(shape=n)
  obj = cp.Maximize(cp.sum(cp.entr(x)))
  constraints = [p_mat @ x <= 0, x >= eps * np.ones(n)]
  constraints.append(cp.sum(x) == 1)
  prob = cp.Problem(obj, constraints)
  prob.solve()
  return x.value.reshape((-1, 1))


def _max_entropy_symmetric_nash_avt(p_mat, num_agents, num_tasks, eps=1e-9):
  """Solves for the maxent symmetric nash for symmetric 2P zero-sum games.

    This covers the agent-vs-task cases.

    Using convex programming:
      min x^Tlog(x) + y^Tlog(y)
      s.t.
      x >= 0
      1^T * x = 1
      y >= 0
      1^T * y = 1
      forall s, such that s has exactly one unit mass on an agent strategy
      and one unit mass on a task strategy,
      s^T*p_mat*z <= 0, where z = [x, y], since game-value is 0.

  Args:
    p_mat: an N*N anti-symmetric payoff matrix for the row player
    num_agents: number of agents
    num_tasks: number of tasks
    eps: minimum probability threshold

  Returns:
    (x*, y*): a maxent symmetric nash
  """
  assert np.array_equal(p_mat, -p_mat.T) and eps >= 0 and eps <= 0.5
  n = len(p_mat)
  assert n == num_agents + num_tasks
  x = cp.Variable(shape=num_agents)
  y = cp.Variable(shape=num_tasks)
  z = cp.hstack([x, y])
  obj = cp.Maximize(cp.sum(cp.entr(z)))
  constraints = [
      x >= eps * np.ones(num_agents),
      cp.sum(x) == 1,
      y >= eps * np.ones(num_tasks),
      cp.sum(y) == 1,
  ]

  dev_payoffs = p_mat @ z
  for a_idx in range(num_agents):
    for t_idx in range(num_tasks):
      pure_strategy = np.zeros(n)
      pure_strategy[a_idx] = 1
      pure_strategy[num_agents + t_idx] = 1
      pure_strategy = pure_strategy.reshape((1, -1))
      constraints.append(pure_strategy @ dev_payoffs <= 0)

  prob = cp.Problem(obj, constraints)
  prob.solve()
  return x.value.reshape((-1, 1)), y.value.reshape((-1, 1))


def nash_averaging_avt_matrix(s_mat, eps=0.0):
  """Apply the agent-vs-task Nash Averaging from Appendix D, from a matrix.

  Args:
    s_mat: The S matrix from the paper, representing m rows (agents) and n
      columns (tasks), with scores for the agent on the task. Note that the
      values need not be normalized, but will be normalized across tasks before
      being processed.
    eps: minimum probability threshold.

  Returns:
    maxent_nash: nash mixture for row player and column player
    nash_avg_score: the expected payoff under maxent_nash
  """
  m, n = s_mat.shape
  min_payoffs = np.min(s_mat, axis=0)
  max_payoffs = np.max(s_mat, axis=0)
  std_p_mat = (s_mat - min_payoffs) / (max_payoffs - min_payoffs)
  a_mat = np.block([
      [np.zeros(shape=(m, m)), std_p_mat],
      [-std_p_mat.T, np.zeros(shape=(n, n))],
  ])
  pa_sol, pe_sol = _max_entropy_symmetric_nash_avt(
      a_mat, num_agents=m, num_tasks=n, eps=eps)
  pa, pe = np.asarray(pa_sol), np.asarray(pe_sol)
  return (pa, pe), (std_p_mat.dot(pe), -std_p_mat.T.dot(pa))


def nash_averaging(game, eps=0.0, a_v_a=True):
  """Nash averaging, see https://arxiv.org/abs/1806.02643.

  Args:
    game: a pyspiel game
    eps: minimum probability mass for maxent nash
    a_v_a: whether it is Agent-vs-Agent or Agent-vs-Task

  Returns:
    maxent_nash: nash mixture for row player and column player
    nash_avg_score: the expected payoff under maxent_nash
  """

  p_mat = game_payoffs_array(game)
  if len(p_mat) != 2:
    raise ValueError("Nash Averaging works only for two players.")
  if np.max(np.abs(p_mat[0] + p_mat[1])) > 0:
    raise ValueError("Must be zero-sum")
  if a_v_a:
    if not np.array_equal(p_mat[0], -p_mat[0].T):
      raise ValueError(
          "AvA only works for symmetric two-player zero-sum games.")
    maxent_nash = np.array(_max_entropy_symmetric_nash(p_mat[0], eps=eps))
    return maxent_nash, p_mat[0].dot(maxent_nash)

  # For AvT, see appendix D of the paper.
  # Here assumes the row player represents agents and the column player
  # represents tasks.
  # game does not have to be symmetric
  return nash_averaging_avt_matrix(p_mat[0], eps=eps)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python example of multiagent Nash Q-learners."""

import enum
import logging
from absl import app

from open_spiel.python import rl_environment
from open_spiel.python.algorithms.tabular_multiagent_qlearner import MAQLearner
from open_spiel.python.algorithms.tabular_multiagent_qlearner import TwoPlayerNashSolver
from open_spiel.python.algorithms.tabular_qlearner import QLearner


class Action(enum.IntEnum):
  STAY = 0
  LEFT = 1
  UP = 2
  RIGHT = 3
  DOWN = 4


def print_iteration(actions, state):
  """Print actions and state."""
  logging.info("Action taken by agent 0: %s", Action(actions[0]).name)
  logging.info("Action taken by agent 1: %s", Action(actions[1]).name)
  logging.info("Board state:\n %s", state)
  logging.info("-" * 80)


def marl_path_finding_example(_):
  """Example usage of multiagent Nash Q-learner.

  Based on https://www.jmlr.org/papers/volume4/hu03a/hu03a.pdf
  """

  logging.info("Creating the Grid Game")
  env = rl_environment.Environment(
      "pathfinding", grid="B.A\n...\na.b", players=2, step_reward=-1.)

  qlearner = QLearner(0, env.game.num_distinct_actions())
  nashqlearner = MAQLearner(1, 2, [env.game.num_distinct_actions()] * 2,
                            TwoPlayerNashSolver())

  time_step = env.reset()
  actions = [None, None]

  while not time_step.last():
    actions = [
        qlearner.step(time_step).action,
        nashqlearner.step(time_step, actions).action
    ]
    time_step = env.step(actions)
    print_iteration(actions, env.get_state)


if __name__ == "__main__":
  app.run(marl_path_finding_example)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A base environment for debate with style actions (logos) and private info.
"""

import dataclasses

from open_spiel.python.games.chat_games.envs.comm_substrates import debates
from open_spiel.python.games.chat_games.envs.utils import header


action_keys = tuple(['style'])
action_defaults = tuple(['logos'])
info_keys = tuple(['info', 'topic'])
info_defaults = tuple(['NA', 'NA'])

w_opts = (debates.W_OPTS_PREFIX +
          'Debate Topic: {topic}\n' +
          'Position: {info}\n' +
          'Style: Make a {style} style argument.' +
          debates.PLAIN)

context = '''You are an intelligent assistant in a debate with another debater.
 The debate topic is given. The goal is to provide arguments that support your
 position as well as arguments against your opponents position. An argument style
is also given. Attempt to craft your arguments according to this given style.

Here are some useful definitions of argument styles:

- logos appeals to the audiences reason, building up logical arguments.

- ethos appeals to the speakers status or authority, making the audience more
likely to trust them.

- pathos appeals to the emotions, trying to make the audience feel angry or
sympathetic, for example.

Try to construct a strong argument to support your position.
'''

HEADER = header.Header(debates.PLAIN,
                       w_opts,
                       debates.strip_msg,
                       debates.SPECIAL_CHARS,
                       action_keys,
                       info_keys,
                       context)


@dataclasses.dataclass(frozen=True)
class Scenario(header.BaseScenario):
  style: str = 'logos'
  topic: str = 'NA'
  info: str = 'NA'



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visualization for single/multi-population dynamics in normal-form games.

  Example:

  game = pyspiel.load_game("matrix_pd")
  payoff_tensor = utils.game_payoffs_array(game)
  dyn = dynamics.MultiPopulationDynamics(payoff_tensor, dynamics.replicator)

  ax = plt.subplot(projection="2x2")
  ax.quiver(dyn)
"""

from absl import logging

# pylint: disable=g-import-not-at-top
try:
  from matplotlib import axes
  from matplotlib import projections
  from matplotlib import transforms
  from matplotlib import font_manager
  from matplotlib import rcParams
  from matplotlib.text import Text
  from matplotlib.path import Path
  from matplotlib.patches import PathPatch
  from matplotlib.patches import FancyArrowPatch
  from matplotlib.collections import LineCollection
  import matplotlib.cm
  import matplotlib.colors
except ImportError as e:
  logging.info("If your tests failed with the error 'ImportError: No module "
               "named functools_lru_cache', this is a known bug in matplotlib "
               "and there is a workaround (run sudo apt install "
               "python-backports.functools-lru-cache. See: "
               "https://github.com/matplotlib/matplotlib/issues/9344.")
  raise ImportError(str(e)) from e

import numpy as np

from open_spiel.python.egt import utils


def _eval_dynamics_2x2_grid(dynamics, num_points):
  """Evaluates dynamics on a 2-D mesh-grid.

  Args:
    dynamics: Population dynamics of type `dynamics.MultiPopulationDynamics`.
    num_points: Number of points along each dimension of the grid.

  Returns:
    Mesh-grid (x, y) and corresponding derivatives of the first action for
      player 1 and 2 (u, v).
  """
  assert dynamics.payoff_tensor.shape == (2, 2, 2)

  x = np.linspace(0., 1., num_points + 2)[1:-1]
  x, y = np.meshgrid(x, x)
  u = np.empty(x.shape)
  v = np.empty(x.shape)

  for i in range(num_points):
    for j in range(num_points):
      row_state = np.array([x[i, j], 1. - x[i, j]])
      col_state = np.array([y[i, j], 1. - y[i, j]])
      state = np.concatenate((row_state, col_state))
      dstate = dynamics(state)
      u[i][j] = dstate[0]
      v[i][j] = dstate[2]
  return x, y, u, v


def _rk12_step(func, y0, dt):
  """Improved Euler-Integration step to integrate dynamics.

  Args:
    func: Function handle to time derivative.
    y0:   Current state.
    dt:   Integration step.

  Returns:
    Next state.
  """
  dy = func(y0)
  y_ = y0 + dt * dy
  return y0 + dt / 2. * (dy + func(y_))


class Dynamics2x2Axes(axes.Axes):
  """Axes for 2x2 game dynamics.

  This class provides plotting functions for dynamics in two-player 2x2 games.

  Attributes:
    name: Used for projection keyword when creating a new axes.
  """
  name = "2x2"

  def cla(self):
    """Clear the current axes."""
    super(Dynamics2x2Axes, self).cla()
    self.set_aspect("equal")
    self.set_xlim(0, 1)
    self.set_ylim(0, 1)

  def quiver(self,
             dynamics,
             num_points=9,
             normalize=False,
             pivot="middle",
             **kwargs):
    """Visualizes the dynamics as a directional field plot.

    Args:
      dynamics: Population dynamics of type `dynamics.MultiPopulationDynamics`.
      num_points: Number of points along each dimension of the plot.
      normalize: Normalize each arrow to unit-length.
      pivot: In `{"tail", "middle", "tip"}`, optional, default: "middle". The
        part of the arrow that is anchored to the X, Y grid. The arrow rotates
        about this point.
      **kwargs: Additional keyword arguments passed on to `Axes.quiver`.

    Returns:
      The `quiver.Quiver` object created by calling `Axes.quiver`.
    """
    x, y, u, v = _eval_dynamics_2x2_grid(dynamics, num_points)

    if normalize:
      norm = np.sqrt(u**2 + v**2)
      u = np.divide(u, norm, out=np.zeros_like(u), where=norm != 0)
      v = np.divide(v, norm, out=np.zeros_like(v), where=norm != 0)

    return super(Dynamics2x2Axes, self).quiver(
        x, y, u, v, pivot=pivot, **kwargs)

  def streamplot(self,
                 dynamics,
                 num_points=50,
                 linewidth=None,
                 color=None,
                 **kwargs):
    """Visualizes the dynamics as a streamline plot.

    Args:
      dynamics: Population dynamics of type `dynamics.MultiPopulationDynamics`.
      num_points: Number of points along each dimension of the plot.
      linewidth: In `{None, float, "velocity"}`, optional, default: None. If
        `linewidth="velocity"`, line width is scaled by the velocity of the
        dynamics. Defaults to `rcParams` if `linewidth=None`.
      color: In `{None, string, (r,g,b), (r,g,b,a), "velocity"}`, default: None.
        If `color="velocity"`, velocity of dynamics is used to color the
        streamlines. Defaults to `rcParams` if `color=None`.
      **kwargs: Additional keyword arguments passed on to `Axes.streamplot`.

    Returns:
      The `streamplot.StreamplotSet` created by calling `Axes.streamplot`.
    """

    x, y, u, v = _eval_dynamics_2x2_grid(dynamics, num_points)

    if linewidth == "velocity" or color == "velocity":
      vel = np.sqrt(u**2 + v**2)
      vel = vel - np.min(vel)
      vel = vel / np.max(vel)

      if linewidth == "velocity":
        linewidth = 3. * vel

      if color == "velocity":
        color = vel

    return super(Dynamics2x2Axes, self).streamplot(
        x, y, u, v, minlength=0.1, linewidth=linewidth, color=color, **kwargs)


projections.register_projection(Dynamics2x2Axes)


class SimplexTransform(transforms.Transform):
  """Affine transform to project the 2-simplex to 2D Cartesian space."""
  input_dims = 3
  output_dims = 2

  _MATRIX = np.array([[0., 0.], [1., 0.], [0.5, np.sqrt(3) / 2.]])

  def transform_affine(self, values):
    return np.matmul(values, SimplexTransform._MATRIX)


class SimplexStreamMask(object):
  """Mask of regular discrete cells to track trajectories/streamlines.

  Also see `matplotlib.streamplot.StreamMask`.
  """

  def __init__(self, density=1.):
    self._n = int(30. * density)
    self._mask = np.zeros([self._n + 1] * 2 + [2], dtype=bool)
    self.shape = self._mask.shape

  def index(self, point):
    """Computes index given a point on the simplex."""
    point = np.array(point)
    idx = np.floor(point[:2] * self._n).astype(int)
    x, y = point[:2] * self._n - idx
    z = int(x + y > 1)
    return tuple(idx.tolist() + [z])

  def point(self, index):
    """Computes point on the simplex given an index."""
    p = np.empty((3,))
    p[0] = (index[0] + (1 + index[2]) / 3.) / float(self._n)
    p[1] = (index[1] + (1 + index[2]) / 3.) / float(self._n)
    p[2] = 1. - p[0] - p[1]
    return p if p[2] > 0. else None

  def __getitem__(self, point):
    return self._mask.__getitem__(self.index(point))

  def __setitem__(self, point, val):
    return self._mask.__setitem__(self.index(point), val)


class Dynamics3x3Axes(axes.Axes):
  """Axes for 3x3 game dynamics.

  This class provides plotting functions for dynamics in symmetric 3x3 games.

  Attributes:
    name: Used for projection keyword when creating a new axes.
  """
  name = "3x3"
  _VERTICES = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]

  def __init__(self, fig, rect, *args, **kwargs):
    self._simplex_transform = SimplexTransform()
    self._labels = None
    super(axes.Axes, self).__init__(fig, rect, *args, **kwargs)

  def cla(self):
    """Clear the current axes."""
    super(axes.Axes, self).cla()
    self.set_aspect("equal")
    self.get_xaxis().set_visible(False)
    self.get_yaxis().set_visible(False)
    self.patch.set_visible(False)
    self.set_frame_on(False)

    # draw invisiple vertices to set x/y limits of plot
    self.scatter(Dynamics3x3Axes._VERTICES, alpha=0.)
    self.margins(0.15)

    self.bgpatch = self._create_bgpatch(
        facecolor=rcParams["axes.facecolor"],
        edgecolor=rcParams["axes.edgecolor"],
        linewidth=rcParams["axes.linewidth"],
        zorder=-1)
    self.add_artist(self.bgpatch)

    if rcParams["axes.grid"]:
      self.grid = self._create_grid(
          color=rcParams["grid.color"],
          alpha=rcParams["grid.alpha"],
          linestyle=rcParams["grid.linestyle"],
          linewidth=rcParams["grid.linewidth"],
          zorder=0)
      self.add_collection(self.grid)

    self.ticks, self.tick_labels = self._create_ticks(
        color=rcParams["xtick.color"], zorder=0)
    self.add_collection(self.ticks)
    for label in self.tick_labels:
      self.add_artist(label)

  def _create_bgpatch(self, **kwargs):
    codes = [Path.MOVETO] + [Path.LINETO] * 2 + [Path.CLOSEPOLY]
    vertices = self._VERTICES + [self._VERTICES[0]]
    vertices = self._simplex_transform.transform(np.array(vertices))
    return PathPatch(Path(vertices, codes), **kwargs)

  def _create_grid(self, step=0.2, **kwargs):
    x = np.arange(step, 1., step)
    n = x.shape[0]
    line_start, line_end = np.zeros((n, 3)), np.zeros((n, 3))
    line_start[:, 0] = line_end[::-1, 1] = x
    line_start[:, 2] = line_end[::-1, 0] = 1. - x
    segs = np.zeros((3 * n, 2, 2))
    for i, perm in enumerate([(0, 2, 1), (1, 0, 2), (2, 1, 0)]):
      start = self._simplex_transform.transform(line_start[:, perm])
      end = self._simplex_transform.transform(line_end[:, perm])
      segs[i * n:(i + 1) * n, 0, :], segs[i * n:(i + 1) * n, 1, :] = start, end
    line_segments = LineCollection(segs, **kwargs)
    return line_segments

  def _create_ticks(self, step=0.2, tick_length=0.025, **kwargs):
    x = np.arange(step, 1., step)
    n = x.shape[0]

    tick_start, tick_end = np.zeros((n, 3)), np.zeros((n, 3))
    tick_start[:, 0] = x
    tick_start[:, 2] = 1. - x
    tick_end[:, 0] = x
    tick_end[:, 2] = 1. - x + tick_length
    tick_end[:, 1] = -tick_length

    tick_labels = []
    ha = ["center", "left", "right"]
    va = ["top", "bottom", "center"]
    rot = [-60, 60, 0]

    segs = np.zeros((n * 3, 2, 2))
    for i, perm in enumerate([(0, 2, 1), (1, 0, 2), (2, 1, 0)]):
      start = self._simplex_transform.transform(tick_start[:, perm])
      end = self._simplex_transform.transform(tick_end[:, perm])
      segs[i * n:(i + 1) * n, 0, :], segs[i * n:(i + 1) * n, 1, :] = start, end

      for j, x_ in enumerate(x):
        tick_labels.append(
            Text(
                end[j, 0],
                end[j, 1],
                "{0:.1f}".format(x_),
                horizontalalignment=ha[i],
                verticalalignment=va[i],
                rotation=rot[i],
                color=kwargs["color"],
                fontsize=rcParams["xtick.labelsize"]))
    line_segments = LineCollection(segs, **kwargs)
    return line_segments, tick_labels

  def _create_labels(self, labels, padding):
    artists = []
    aligns = ["top", "top", "bottom"]
    for label, pos, align in zip(labels, self._VERTICES, aligns):
      x, y = self._simplex_transform.transform(pos)
      labelpad = padding if align == "bottom" else -padding
      label = Text(
          x=x,
          y=y + labelpad,
          text=label,
          fontproperties=font_manager.FontProperties(
              size=rcParams["axes.labelsize"],
              weight=rcParams["axes.labelweight"]),
          color=rcParams["axes.labelcolor"],
          verticalalignment=align,
          horizontalalignment="center")
      artists.append(label)
    return artists

  def get_labels(self):
    return self._labels

  def set_labels(self, labels, padding=0.02):
    assert len(labels) == 3
    if self._labels is None:
      self._labels = self._create_labels(labels, padding)
      for label in self._labels:
        self.add_artist(label)
    else:
      for artist, label in zip(self._labels, labels):
        artist.set_text(label)

  labels = property(get_labels, set_labels)

  def can_zoom(self):
    return False

  def can_pan(self):
    return False

  def plot(self, points, **kwargs):
    """Creates a line plot.

    Args:
      points: Points in policy space.
      **kwargs: Additional keyword arguments passed on to `Axes.plot`.

    Returns:
      The line plot.
    """
    points = np.array(points)
    assert points.shape[1] == 3
    points = self._simplex_transform.transform(points)
    return super(Dynamics3x3Axes, self).plot(points[:, 0], points[:, 1],
                                             **kwargs)

  def scatter(self, points, **kwargs):
    """Creates a scatter plot.

    Args:
      points: Points in policy space.
      **kwargs: Additional keyword arguments passed on to `Axes.scatter`.

    Returns:
      The scatter plot.
    """
    points = np.array(points)
    assert points.shape[1] == 3
    points = self._simplex_transform.transform(points)
    return super(Dynamics3x3Axes, self).scatter(points[:, 0], points[:, 1],
                                                **kwargs)

  def quiver(self,
             dynamics,
             step=0.05,
             boundary=False,
             normalize=False,
             pivot="middle",
             **kwargs):
    """Visualizes the dynamics as a directional field plot.

    Args:
      dynamics: Population dynamics of type `dynamics.SinglePopulationDynamics`.
      step: Distance between arrows along one dimension.
      boundary: Include arrows on the boundary/face of the simplex.
      normalize: Normalize each arrow to unit-length.
      pivot: In `{"tail", "middle", "tip"}`, optional, default: "middle". The
        part of the arrow that is anchored to the X, Y grid. The arrow rotates
        about this point.
      **kwargs: Additional keyword arguments passed on to `Axes.quiver`.

    Returns:
      The `quiver.Quiver` object created by calling `Axes.quiver`.
    """
    x = np.array([x for x in utils.grid_simplex(step=step, boundary=boundary)])
    dx = np.apply_along_axis(dynamics, 1, x)

    p = self._simplex_transform.transform(x)
    v = self._simplex_transform.transform(dx)

    x, y = p[:, 0], p[:, 1]
    u, v = v[:, 0], v[:, 1]

    if normalize:
      norm = np.sqrt(u**2 + v**2)
      u, v = u / norm, v / norm

    if "pivot" not in kwargs:
      kwargs["pivot"] = "middle"

    return super(Dynamics3x3Axes, self).quiver(x, y, u, v, **kwargs)

  def _linecollection(self, points, linewidth, color):
    points = self._simplex_transform.transform(points).reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    lc = LineCollection(segments, linewidths=linewidth, color=color)
    return lc

  def _integrate(self, x, func, mask, dt, min_dist=0.01):
    cells = []
    trajectory = [x]
    x_ = x
    for dt in [dt, -dt]:
      while not mask[x]:
        cell = mask.index(x)
        cells.append(cell)
        while mask.index(x) == cell:
          # integrate up to cell boundary
          if np.sqrt(np.sum((x_ - x)**2)) > min_dist:
            x_ = x
            if dt > 0:
              trajectory.append(x)
            else:
              trajectory.insert(0, x)

          x = _rk12_step(func, x, dt=dt)

        if dt > 0:
          mask[trajectory[-1]] = True
        else:
          mask[trajectory[0]] = True

      # restore to integrate backwards
      if dt > 0. and len(cells):
        trajectory.append(_rk12_step(func, x, dt=-dt))
        mask[mask.point(cells[0])] = False
        x = trajectory[0]
        x_ = x
      else:
        trajectory.insert(0, _rk12_step(func, x, dt=-dt))
    return (np.array(trajectory), cells) if len(trajectory) > 2 else None

  def streamplot(self,
                 dynamics,
                 initial_points=None,
                 dt=0.01,
                 density=1.,
                 min_length=0.4,
                 linewidth=None,
                 color="k",
                 **kwargs):
    """Visualizes the dynamics as a streamline plot.

    Mimics the visuals of `Axes.streamplot` for simplex plots.

    Args:
      dynamics: Population dynamics of type `dynamics.SinglePopulationDynamics`.
      initial_points: Starting points for streamlines
      dt: Integration step.
      density: Controls the density of streamlines in the plot.
      min_length: Streamlines with length < min_length will be discarded.
      linewidth: In `{None, float, "velocity"}`, optional, default: None. If
        `linewidth="velocity"`, line width is scaled by the velocity of the
        dynamics. Defaults to `rcParams` if `linewidth=None`.
      color: In `{None, string, (r,g,b), (r,g,b,a), "velocity"}`, default: None.
        If `color="velocity"`, velocity of dynamics is used to color the
        streamlines. Defaults to `rcParams` if `color=None`.
      **kwargs: Additional keyword arguments passed on to `Axes.streamplot`.

    Returns:
      The `SimplexStreamMask`.
    """
    mask = SimplexStreamMask(density=density)
    trajectories = []

    if initial_points is None:
      eps = 0.1
      initial_points = np.array([[1. - eps, eps / 2., eps / 2.],
                                 [eps / 2., 1. - eps, eps / 2.],
                                 [eps / 2., eps / 2., 1. - eps]])
      initial_points = np.vstack(
          (initial_points, utils.sample_from_simplex(100)))
      # TODO(author10): add heuristic for initial points

    else:
      initial_points = np.array(initial_points)
      assert initial_points.ndim == 2
      assert initial_points.shape[1] == 3

    # generate trajectories
    for p in initial_points:
      # center initial point on grid cell
      p = mask.point(mask.index(p))
      res = self._integrate(p, dynamics, mask, dt=dt)
      if res is not None:
        t, cells = res  # pylint: disable=unpacking-non-sequence
        cum_len = np.cumsum(
            np.sqrt(
                np.diff(t[:, 0])**2 + np.diff(t[:, 1])**2 +
                np.diff(t[:, 2])**2))
        if cum_len[-1] < min_length:
          for cell in cells:
            mask[mask.point(cell)] = False
          continue
        trajectories.append(t)

    lc_color = arrow_color = color
    lc_linewidth = linewidth

    if linewidth == "velocity" or color == "velocity":
      vel_max = 0
      vel_min = float("inf")
      velocities = []
      for t in trajectories:
        dx = np.apply_along_axis(dynamics, 1, t)
        vel = np.sqrt(np.sum(dx**2, axis=1))
        vel_max = max(np.max(vel), vel_max)
        vel_min = min(np.min(vel), vel_min)
        velocities.append(vel)

    # add trajectories to plot
    for i, t in enumerate(trajectories):
      cum_len = np.cumsum(
          np.sqrt(
              np.diff(t[:, 0])**2 + np.diff(t[:, 1])**2 + np.diff(t[:, 2])**2))
      mid_idx = np.searchsorted(cum_len, cum_len[-1] / 2.)

      if linewidth == "velocity" or color == "velocity":
        vel = (velocities[i] - vel_min) / vel_max

        if linewidth == "velocity":
          lc_linewidth = 3. * vel + 0.5

        if color == "velocity":
          cmap = matplotlib.cm.get_cmap(rcParams["image.cmap"])
          lc_color = cmap(vel)
          arrow_color = cmap(vel[mid_idx])

      lc = self._linecollection(t, linewidth=lc_linewidth, color=lc_color)
      self.add_collection(lc)

      # add arrow centered on trajectory
      arrow_tail = self._simplex_transform.transform(t[mid_idx - 1])
      arrow_head = self._simplex_transform.transform(t[mid_idx])
      arrow_kw = dict(arrowstyle="-|>", mutation_scale=10 * 1.)
      arrow_patch = FancyArrowPatch(
          arrow_tail,
          arrow_head,
          linewidth=None,
          color=arrow_color,
          zorder=3,
          **arrow_kw)
      self.add_patch(arrow_patch)
    return mask


projections.register_projection(Dynamics3x3Axes)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for open_spiel.python.algorithms.tabular_multiagent_qlearner."""

from absl.testing import absltest
import numpy as np

from open_spiel.python import rl_environment
from open_spiel.python.algorithms.tabular_qlearner import QLearner
from open_spiel.python.algorithms.wolf_phc import WoLFPHC

SEED = 18763511


class WoLFTest(absltest.TestCase):

  def test_simple_pathfinding_run(self):
    env = rl_environment.Environment(
        "pathfinding", grid="B.A\n...\na.b", players=2, step_reward=-1.)

    with self.subTest("wolf_phc"):
      qlearner = QLearner(0, env.game.num_distinct_actions())
      wolflearner = WoLFPHC(1, env.game.num_distinct_actions())
      time_step = env.reset()
      step_cnt = 0

      while not time_step.last():
        actions = [
            qlearner.step(time_step).action,
            wolflearner.step(time_step).action
        ]
        time_step = env.step(actions)
        step_cnt += 1

      self.assertLess(step_cnt, 500)

  def test_rps_run(self):
    env = rl_environment.Environment("matrix_rps")
    wolf0 = WoLFPHC(0, env.game.num_distinct_actions())
    wolf1 = WoLFPHC(1, env.game.num_distinct_actions())

    for _ in range(1000):
      time_step = env.reset()
      actions = [wolf0.step(time_step).action, wolf1.step(time_step).action]
      time_step = env.step(actions)
      wolf0.step(time_step)
      wolf1.step(time_step)

    with self.subTest("correct_rps_strategy"):
      time_step = env.reset()
      learner0_strategy, learner1_strategy = wolf0.step(
          time_step).probs, wolf1.step(time_step).probs
      np.testing.assert_array_almost_equal(
          np.asarray([1 / 3, 1 / 3, 1 / 3]),
          learner0_strategy.reshape(-1),
          decimal=4)
      np.testing.assert_array_almost_equal(
          np.asarray([1 / 3, 1 / 3, 1 / 3]),
          learner1_strategy.reshape(-1),
          decimal=4)


if __name__ == "__main__":
  np.random.seed(SEED)
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""IMPALA agent implemented in JAX.

This is a basic IMPALA agent adapted from the example in the Haiku project:
https://github.com/deepmind/dm-haiku/tree/main/examples/impala
"""

import collections
import functools
from typing import Any, Callable, Dict, Optional, Tuple

import chex
import haiku as hk
import jax
from jax.example_libraries import optimizers
import jax.numpy as jnp
import numpy as np
import optax
import rlax
import tree

from open_spiel.python import rl_agent
from open_spiel.python.examples.rrps_poprl import rl_environment


AgentOutput = collections.namedtuple(
    "AgentOutput", ["policy_logits", "values", "action", "prediction_logits"]
)
NetOutput = collections.namedtuple(
    "NetOutput", ["policy_logits", "value", "prediction_logits"]
)
Transition = collections.namedtuple(
    "Transition", ["timestep", "agent_out", "agent_state"]
)
NetFactory = Callable[[int], hk.RNNCore]
Nest = Any


# The IMPALA paper sums losses, rather than taking the mean.
# We wrap rlax to do so as well.
def policy_gradient_loss(logits, *args):
  """rlax.policy_gradient_loss, but with sum(loss) and [T, B, ...] inputs."""
  # jax.experimental.host_callback.id_print(logits.shape)
  # print(logits.shape)
  mean_per_batch = jax.vmap(rlax.policy_gradient_loss, in_axes=1)(logits, *args)
  total_loss_per_batch = mean_per_batch * logits.shape[0]
  return jnp.sum(total_loss_per_batch)


def entropy_loss(logits, *args):
  """rlax.entropy_loss, but with sum(loss) and [T, B, ...] inputs."""
  mean_per_batch = jax.vmap(rlax.entropy_loss, in_axes=1)(logits, *args)
  total_loss_per_batch = mean_per_batch * logits.shape[0]
  return jnp.sum(total_loss_per_batch)


def mean_pred_loss_without_batch(
    logits_t: chex.Array,
    labels: chex.Array,
) -> chex.Array:
  """Mean prediction loss without batch dimension."""
  chex.assert_rank([logits_t, labels], [2, 1])
  chex.assert_type([logits_t, labels], [float, int])
  labels_one_hot = jax.nn.one_hot(labels, logits_t.shape[-1])
  softmax_xent = -jnp.sum(labels_one_hot * jax.nn.log_softmax(logits_t))
  softmax_xent /= labels.shape[0]
  return softmax_xent


def prediction_loss(logits, labels):
  # print(logits.shape)  -> [T, B, num_preds]
  # print(labels.shape)  -> [T, B]
  mean_per_batch = jax.vmap(mean_pred_loss_without_batch, in_axes=1)(
      logits, labels
  )
  total_loss_per_batch = mean_per_batch * logits.shape[0]
  return jnp.sum(total_loss_per_batch)


def _preprocess_none(t) -> np.ndarray:
  if t is None:
    return np.array(0.0, dtype=np.float32)
  else:
    return np.asarray(t)


def preprocess_step(
    timestep: rl_environment.TimeStep, num_players
) -> rl_environment.TimeStep:
  # TODO(author5): fix for our time steps (should be multiple discounts)
  if timestep.discounts is None:
    timestep = timestep._replace(discounts=[1.0] * num_players)
  if timestep.rewards is None:
    timestep = timestep._replace(rewards=[0.0] * num_players)
  # print(timestep)
  return tree.map_structure(_preprocess_none, timestep)


# dm_env: return TimeStep(StepType.FIRST, None, None, observation)
# OpenSpiel: "observations", "rewards", "discounts", "step_type"
def restart(dummy_obs, num_players):
  all_obs = {
      "info_state": [dummy_obs.copy() for i in range(num_players)],
      "legal_actions": [np.zeros(3)],
      "prediction_label": 0,
  }
  return rl_environment.TimeStep(
      all_obs, [0.0], None, rl_environment.StepType.FIRST
  )


class BasicRNN(hk.RNNCore):
  """A simple recurrent neural network."""

  def __init__(
      self,
      player_id,
      num_actions,
      hidden_layer_sizes,
      num_predictions,
      name=None,
  ):
    super().__init__(name=name)
    self._player_id = player_id
    self._num_actions = num_actions
    self._num_predictions = num_predictions
    self._hidden_layer_sizes = hidden_layer_sizes
    if isinstance(hidden_layer_sizes, int):
      self._hidden_layer_sizes = [hidden_layer_sizes]
    elif isinstance(hidden_layer_sizes, tuple):
      self._hidden_layer_sizes = list(hidden_layer_sizes)
    self._core = hk.ResetCore(hk.LSTM(256))

  def initial_state(self, batch_size):
    return self._core.initial_state(batch_size)

  def __call__(self, x: rl_environment.TimeStep, state):
    x = jax.tree_util.tree_map(lambda t: t[None, ...], x)
    return self.unroll(x, state)

  def unroll(self, x, state):
    modules = [hk.Flatten()]
    for hsize in self._hidden_layer_sizes:
      modules.append(hk.Linear(hsize))
      modules.append(jax.nn.relu)
    torso_net = hk.Sequential(modules)
    torso_output = hk.BatchApply(torso_net)(
        x.observations["info_state"][self._player_id]
    )
    should_reset = jnp.equal(x.step_type, int(rl_environment.StepType.FIRST))
    core_input = (torso_output, should_reset)
    core_output, state = hk.dynamic_unroll(self._core, core_input, state)
    policy_logits = hk.Linear(self._num_actions)(core_output)
    prediction_logits = hk.Linear(self._num_predictions)(core_output)
    value = hk.Linear(1)(core_output)
    value = jnp.squeeze(value, axis=-1)
    return (
        NetOutput(
            policy_logits=policy_logits,
            value=value,
            prediction_logits=prediction_logits,
        ),
        state,
    )
    # torso_output = torso_net(x.observations["info_state"][self._player_id])
    # policy_logits = hk.Linear(self._num_actions)(torso_output)
    # prediction_logits = hk.Linear(self._num_predictions)(torso_output)
    # value = hk.Linear(1)(torso_output)
    # value = jnp.squeeze(value, axis=-1)
    # return NetOutput(policy_logits=policy_logits,
    #                  value=value,
    #                  prediction_logits=prediction_logits), state


class IMPALA(rl_agent.AbstractAgent):
  """IMPALA agent implementation in JAX."""

  def __init__(
      self,
      player_id,
      state_representation_size,
      num_actions,
      num_players,
      unroll_len,
      net_factory: NetFactory,
      rng_key,
      max_abs_reward,
      learning_rate=0.0001,
      entropy=0.01,
      discount_factor=0.99,
      hidden_layers_sizes=128,
      batch_size=16,
      num_predictions=10,
      prediction_weight=0.01,
      max_global_gradient_norm=None,
  ):
    self._player_id = player_id
    self._state_representation_size = state_representation_size
    self._num_actions = num_actions
    self._num_players = num_players
    self._unroll_len = unroll_len
    self._rng_key = rng_key
    self._max_abs_reward = max_abs_reward
    self._learning_rate = learning_rate
    self._batch_size = batch_size
    self._discount_factor = discount_factor
    self._entropy = entropy
    self._hidden_layer_sizes = hidden_layers_sizes
    self._num_predictions = num_predictions
    self._prediction_weight = prediction_weight
    self._dummy_obs = np.zeros(
        shape=state_representation_size, dtype=np.float32
    )

    # pylint: disable=too-many-function-args
    net_factory = functools.partial(
        net_factory,
        player_id,
        num_actions,
        hidden_layers_sizes,
        num_predictions,
    )
    # Instantiate two hk.transforms() - one for getting the initial state of the
    # agent, another for actually initializing and running the agent.
    _, self._initial_state_apply_fn = hk.without_apply_rng(
        hk.transform(lambda batch_size: net_factory().initial_state(batch_size))
    )

    self._init_fn, self._apply_fn = hk.without_apply_rng(
        hk.transform(lambda obs, state: net_factory().unroll(obs, state))
    )

    # Learner components
    # self._opt = optax.rmsprop(5e-3, decay=0.99, eps=1e-7)
    self._opt = optax.rmsprop(self._learning_rate, decay=0.99, eps=1e-7)
    # self._opt = optax.sgd(self._learning_rate)

    # Prepare parameters and initial state
    self._rng_key, subkey = jax.random.split(self._rng_key)
    init_params = self.initial_params(subkey)
    self._frame_count_and_params = (0, jax.device_get(init_params))
    (_, params) = self._frame_count_and_params
    self._opt_state = self._opt.init(params)
    self._rng_key, _ = jax.random.split(self._rng_key)
    self._agent_state = self._agent_state = self.initial_state(None)
    self._traj = []
    self._batch = []
    self._last_policy = None
    self._last_predictions = None

  @functools.partial(jax.jit, static_argnums=0)
  def initial_params(self, rng_key):
    """Initializes the agent params given the RNG key."""

    dummy_inputs = jax.tree_util.tree_map(
        lambda t: np.zeros(t.shape, t.dtype), self._dummy_obs
    )
    dummy_inputs = preprocess_step(
        restart(dummy_inputs, self._num_players), self._num_players
    )
    # Add time and batch dimensions
    dummy_inputs = jax.tree_util.tree_map(
        lambda t: t[None, None, ...], dummy_inputs
    )
    # print(dummy_inputs)
    return self._init_fn(rng_key, dummy_inputs, self.initial_state(1))

  @functools.partial(jax.jit, static_argnums=(0, 1))
  def initial_state(self, batch_size: Optional[int]):
    """Returns agent initial state."""
    # We expect that generating the initial_state does not require parameters.
    return self._initial_state_apply_fn(None, batch_size)

  @functools.partial(jax.jit, static_argnums=(0,))
  def internal_step(
      self,
      rng_key,
      params: hk.Params,
      timestep: rl_environment.TimeStep,
      state: Nest,
  ) -> Tuple[AgentOutput, Nest]:
    """For a given single-step, unbatched timestep, output the chosen action."""
    # Pad timestep, state to be [T, B, ...] and [B, ...] respectively.
    # print("calling internal_step")
    timestep = jax.tree_util.tree_map(lambda t: t[None, None, ...], timestep)
    state = jax.tree_util.tree_map(lambda t: t[None, ...], state)
    net_out, next_state = self._apply_fn(params, timestep, state)
    # print(timestep)
    # Remove the padding from above.
    net_out = jax.tree_util.tree_map(
        lambda t: jnp.squeeze(t, axis=(0, 1)), net_out
    )
    next_state = jax.tree_util.tree_map(
        lambda t: jnp.squeeze(t, axis=0), next_state
    )
    # Sample an action and return.
    action = hk.multinomial(rng_key, net_out.policy_logits, num_samples=1)
    action = jnp.squeeze(action, axis=-1)
    return (
        AgentOutput(
            net_out.policy_logits,
            net_out.value,
            action,
            net_out.prediction_logits,
        ),
        next_state,
    )

  def unroll(
      self,
      params: hk.Params,
      trajectory: rl_environment.TimeStep,
      state: Nest,
  ) -> AgentOutput:
    """Unroll the agent along trajectory."""
    net_out, _ = self._apply_fn(params, trajectory, state)
    return AgentOutput(
        net_out.policy_logits,
        net_out.value,
        action=[],
        prediction_logits=net_out.prediction_logits,
    )

  def _loss(
      self,
      theta: hk.Params,
      trajectories: Transition,
  ) -> Tuple[jnp.ndarray, Dict[str, jnp.ndarray]]:
    """Compute vtrace-based actor-critic loss."""
    # All the individual components are vectorized to be [T, B, ...]
    # print(trajectories)
    # Transition(timestep=TimeStep(observations={
    # 'current_player': array([[-2, -2, -2, -2, -2],
    #   [-2, -2, -2, -2, -2],
    #   [-2, -2, -2, -2, -2],
    #   [-2, -2, -2, -2, -2],
    # Since prediction_label is a scalar, it ends up being [T, B]

    initial_state = jax.tree_util.tree_map(
        lambda t: t[0], trajectories.agent_state
    )
    learner_outputs = self.unroll(theta, trajectories.timestep, initial_state)
    v_t = learner_outputs.values[1:]
    # Remove bootstrap timestep from non-timesteps.
    _, actor_out, _ = jax.tree_util.tree_map(lambda t: t[:-1], trajectories)
    learner_outputs = jax.tree_util.tree_map(lambda t: t[:-1], learner_outputs)
    v_tm1 = learner_outputs.values

    # Get the discount, reward, step_type from the *next* timestep.
    timestep = jax.tree_util.tree_map(lambda t: t[1:], trajectories.timestep)
    discounts = timestep.discounts[self._player_id] * self._discount_factor
    rewards = timestep.rewards[self._player_id]
    if self._max_abs_reward > 0:
      rewards = jnp.clip(rewards, -self._max_abs_reward, self._max_abs_reward)

    # The step is uninteresting if we transitioned LAST -> FIRST.
    # timestep corresponds to the *next* time step, so we filter for FIRST.
    mask = jnp.not_equal(timestep.step_type, int(rl_environment.StepType.FIRST))
    mask = mask.astype(jnp.float32)

    rhos = rlax.categorical_importance_sampling_ratios(
        learner_outputs.policy_logits, actor_out.policy_logits, actor_out.action
    )
    # vmap vtrace_td_error_and_advantage to take/return [T, B, ...].
    vtrace_td_error_and_advantage = jax.vmap(
        rlax.vtrace_td_error_and_advantage, in_axes=1, out_axes=1
    )

    vtrace_returns = vtrace_td_error_and_advantage(
        v_tm1, v_t, rewards, discounts, rhos
    )
    pg_advs = vtrace_returns.pg_advantage
    # print(learner_outputs.policy_logits.shape)
    # jax.experimental.host_callback.id_print(learner_outputs.policy_logits.shape)
    pg_loss = policy_gradient_loss(
        learner_outputs.policy_logits, actor_out.action, pg_advs, mask
    )

    baseline_loss = 0.5 * jnp.sum(jnp.square(vtrace_returns.errors) * mask)
    ent_loss = entropy_loss(learner_outputs.policy_logits, mask)

    pred_loss = 0
    if self._prediction_weight > 0:
      pred_loss = prediction_loss(
          learner_outputs.prediction_logits,
          trajectories.timestep.observations["prediction_label"][:-1],
      )

    total_loss = pg_loss
    total_loss += 0.5 * baseline_loss
    total_loss += self._entropy * ent_loss
    total_loss += self._prediction_weight * pred_loss

    logs = {}
    logs["PG_loss"] = pg_loss
    logs["baseline_loss"] = baseline_loss
    logs["entropy_loss"] = ent_loss
    logs["prediction_loss"] = pred_loss
    logs["total_loss"] = total_loss
    return total_loss, logs

  @functools.partial(jax.jit, static_argnums=0)
  def update(self, params, opt_state, batch: Transition):
    """The actual update function."""
    (_, logs), grads = jax.value_and_grad(self._loss, has_aux=True)(
        params, batch
    )

    grad_norm_unclipped = optimizers.l2_norm(grads)
    updates, updated_opt_state = self._opt.update(grads, opt_state)
    params = optax.apply_updates(params, updates)
    weight_norm = optimizers.l2_norm(params)
    logs.update({
        "grad_norm_unclipped": grad_norm_unclipped,
        "weight_norm": weight_norm,
    })
    return params, updated_opt_state, logs

  def _learning_step(self):
    # print("Learning!!!")
    # Prepare for consumption, then put batch onto device.
    # stacked_batch = jax.tree_multimap(lambda *xs: np.stack(xs, axis=1),
    #                                 *self._batch)
    stacked_batch = jax.tree_util.tree_map(
        lambda *xs: np.stack(xs, axis=1), *self._batch
    )
    # self._device_q.put(jax.device_put(stacked_batch))
    jax.device_put(stacked_batch)
    num_frames, params = self._frame_count_and_params
    params, self._opt_state, _ = self.update(
        params, self._opt_state, stacked_batch
    )
    self._frame_count_and_params = (num_frames + 1, params)
    self._batch = []

  def last_policy(self):
    return self._last_policy

  def last_predictions(self):
    return self._last_predictions

  def step(self, time_step, is_evaluation=False):
    # Hack to run with environments that include the serialized state: simply
    # remove it.
    if "serialized_state" in time_step.observations:
      del time_step.observations["serialized_state"]
    # OpenSpiel time steps have lists of floats. First convert to numpy.
    for p in range(self._num_players):
      time_step.observations["info_state"][p] = np.asarray(
          time_step.observations["info_state"][p]
      )

    # print(time_step)
    # TODO(author5): the arrays need to be the same shape, so when the
    # legal actions are empty vs full, this is a problem.
    # Fix later. for now, replace with a constant
    time_step.observations["legal_actions"] = [np.ones(3)]

    agent_state = self._agent_state
    (_, params) = self._frame_count_and_params
    jax.device_put(params)
    time_step = preprocess_step(time_step, self._num_players)
    self._rng_key, subkey = jax.random.split(self._rng_key)
    agent_out, next_state = self.internal_step(
        subkey, params, time_step, agent_state
    )

    self._last_policy = jax.nn.softmax(agent_out.policy_logits).copy()
    self._last_predictions = jax.nn.softmax(agent_out.prediction_logits).copy()

    transition = Transition(
        timestep=time_step, agent_out=agent_out, agent_state=agent_state
    )
    self._agent_state = next_state

    # Do not add to trajectory or check for learning during evaluation.
    if not is_evaluation:
      self._traj.append(transition)
      # Check for learning step.
      if len(self._traj) >= self._unroll_len:
        trajectory = jax.device_get(self._traj)
        # trajectory = jax.tree_multimap(lambda *xs: np.stack(xs), *trajectory)
        trajectory = jax.tree_util.tree_map(
            lambda *xs: np.stack(xs), *trajectory
        )
        self._batch.append(trajectory)
        self._traj = self._traj[-1:]
        if len(self._batch) >= self._batch_size:
          self._learning_step()

    if time_step.last():
      return None

    assert 0 <= agent_out.action < self._num_actions
    # TODO(author5): get probs from the policy in internal_step
    probs = np.zeros(self._num_actions, dtype=np.float32)
    probs[agent_out.action] = 1.0
    return rl_agent.StepOutput(action=agent_out.action, probs=probs)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for open_spiel.python.algorithms.regret_matching."""

from absl.testing import absltest
import numpy as np

from open_spiel.python.algorithms import regret_matching
from open_spiel.python.egt.utils import game_payoffs_array
import pyspiel


class RegretMatchingTest(absltest.TestCase):

  def test_two_players(self):
    test_a = np.array([[2, 1, 0], [0, -1, -2]])
    test_b = np.array([[2, 1, 0], [0, -1, -2]])

    strategies = regret_matching.regret_matching(
        [test_a, test_b],
        initial_strategies=None,
        iterations=50000,
        prd_gamma=1e-8,
        average_over_last_n_strategies=10)

    self.assertLen(strategies, 2, "Wrong strategy length.")
    self.assertGreater(strategies[0][0], 0.999,
                       "Regret matching failed in trivial case.")

  def test_three_players(self):
    test_a = np.array([[[2, 1, 0], [1, 0, -1]], [[1, 0, -1], [0, -1, -2]]])
    test_b = np.array([[[2, 1, 0], [1, 0, -1]], [[1, 0, -1], [0, -1, -2]]])
    test_c = np.array([[[2, 1, 0], [1, 0, -1]], [[1, 0, -1], [0, -1, -2]]])

    strategies = regret_matching.regret_matching(
        [test_a, test_b, test_c],
        initial_strategies=None,
        iterations=50000,
        gamma=1e-6,
        average_over_last_n_strategies=10)
    self.assertLen(strategies, 3, "Wrong strategy length.")
    self.assertGreater(strategies[0][0], 0.999,
                       "Regret matching failed in trivial case.")

  def test_rps(self):
    game = pyspiel.load_game("matrix_rps")
    payoffs_array = game_payoffs_array(game)
    strategies = regret_matching.regret_matching(
        [payoffs_array[0], payoffs_array[1]],
        initial_strategies=[
            np.array([0.1, 0.4, 0.5]),
            np.array([0.9, 0.1, 0.01])
        ],
        iterations=50000,
        gamma=1e-6)
    self.assertLen(strategies, 2, "Wrong strategy length.")
    # places=1 corresponds to an absolute difference of < 0.001
    self.assertAlmostEqual(strategies[0][0], 1 / 3., places=2)
    self.assertAlmostEqual(strategies[0][1], 1 / 3., places=2)
    self.assertAlmostEqual(strategies[0][2], 1 / 3., places=2)

  def test_biased_rps(self):
    game = pyspiel.load_game("matrix_brps")
    payoffs_array = game_payoffs_array(game)
    strategies = regret_matching.regret_matching(
        [payoffs_array[0], payoffs_array[1]], iterations=50000, gamma=1e-8)
    self.assertLen(strategies, 2, "Wrong strategy length.")
    # places=1 corresponds to an absolute difference of < 0.01
    self.assertAlmostEqual(strategies[0][0], 1 / 16., places=1)
    self.assertAlmostEqual(strategies[0][1], 10 / 16., places=1)
    self.assertAlmostEqual(strategies[0][2], 5 / 16., places=1)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An example of use of PPO.

Note: code adapted (with permission) from
https://github.com/vwxyzjn/cleanrl/blob/master/cleanrl/ppo.py and
https://github.com/vwxyzjn/ppo-implementation-details/blob/main/ppo_atari.py
"""

# pylint: disable=g-importing-member
import collections
from datetime import datetime
import logging
import os
import random
import sys
import time
from absl import app
from absl import flags
import numpy as np
import pandas as pd
import torch
from torch.utils.tensorboard import SummaryWriter

import pyspiel
from open_spiel.python.pytorch.ppo import PPO
from open_spiel.python.pytorch.ppo import PPOAgent
from open_spiel.python.pytorch.ppo import PPOAtariAgent
from open_spiel.python.rl_environment import ChanceEventSampler
from open_spiel.python.rl_environment import Environment
from open_spiel.python.rl_environment import ObservationType
from open_spiel.python.vector_env import SyncVectorEnv


FLAGS = flags.FLAGS

flags.DEFINE_string("exp_name",
                    os.path.basename(__file__).rstrip(".py"),
                    "the name of this experiment")
flags.DEFINE_string("game_name", "atari", "the id of the OpenSpiel game")
flags.DEFINE_float("learning_rate", 2.5e-4,
                   "the learning rate of the optimizer")
flags.DEFINE_integer("seed", 1, "seed of the experiment")
flags.DEFINE_integer("total_timesteps", 10_000_000,
                     "total timesteps of the experiments")
flags.DEFINE_integer("eval_every", 10, "evaluate the policy every N updates")
flags.DEFINE_bool("torch_deterministic", True,
                  "if toggled, `torch.backends.cudnn.deterministic=False`")
flags.DEFINE_bool("cuda", True, "if toggled, cuda will be enabled by default")

# Atari specific arguments
flags.DEFINE_string("gym_id", "BreakoutNoFrameskip-v4",
                    "the id of the environment")
flags.DEFINE_bool(
    "capture_video", False,
    "whether to capture videos of the agent performances (check out `videos` folder)"
)

# Algorithm specific arguments
flags.DEFINE_integer("num_envs", 8, "the number of parallel game environments")
flags.DEFINE_integer(
    "num_steps", 128,
    "the number of steps to run in each environment per policy rollout")
flags.DEFINE_bool(
    "anneal_lr", True,
    "Toggle learning rate annealing for policy and value networks")
flags.DEFINE_bool("gae", True, "Use GAE for advantage computation")
flags.DEFINE_float("gamma", 0.99, "the discount factor gamma")
flags.DEFINE_float("gae_lambda", 0.95,
                   "the lambda for the general advantage estimation")
flags.DEFINE_integer("num_minibatches", 4, "the number of mini-batches")
flags.DEFINE_integer("update_epochs", 4, "the K epochs to update the policy")
flags.DEFINE_bool("norm_adv", True, "Toggles advantages normalization")
flags.DEFINE_float("clip_coef", 0.1, "the surrogate clipping coefficient")
flags.DEFINE_bool(
    "clip_vloss", True,
    "Toggles whether or not to use a clipped loss for the value function, as per the paper"
)
flags.DEFINE_float("ent_coef", 0.01, "coefficient of the entropy")
flags.DEFINE_float("vf_coef", 0.5, "coefficient of the value function")
flags.DEFINE_float("max_grad_norm", 0.5,
                   "the maximum norm for the gradient clipping")
flags.DEFINE_float("target_kl", None, "the target KL divergence threshold")


def setup_logging():
  root = logging.getLogger()
  root.setLevel(logging.DEBUG)

  handler = logging.StreamHandler(sys.stdout)
  handler.setLevel(logging.DEBUG)
  formatter = logging.Formatter(
      "%(asctime)s - %(name)s - %(levelname)s - %(message)s")
  handler.setFormatter(formatter)
  root.addHandler(handler)


def make_single_atari_env(gym_id,
                          seed,
                          idx,
                          capture_video,
                          run_name,
                          use_episodic_life_env=True):
  """Make the single-agent Atari environment."""

  def gen_env():
    game = pyspiel.load_game(
        "atari", {
            "gym_id": gym_id,
            "seed": seed,
            "idx": idx,
            "capture_video": capture_video,
            "run_name": run_name,
            "use_episodic_life_env": use_episodic_life_env
        })
    return Environment(
        game,
        chance_event_sampler=ChanceEventSampler(seed=seed),
        observation_type=ObservationType.OBSERVATION)

  return gen_env


def make_single_env(game_name, seed):

  def gen_env():
    game = pyspiel.load_game(game_name)
    return Environment(game, chance_event_sampler=ChanceEventSampler(seed=seed))

  return gen_env


def main(_):
  setup_logging()

  batch_size = int(FLAGS.num_envs * FLAGS.num_steps)

  if FLAGS.game_name == "atari":
    # pylint: disable=unused-import
    # pylint: disable=g-import-not-at-top
    import open_spiel.python.games.atari

  current_day = datetime.now().strftime("%d")
  current_month_text = datetime.now().strftime("%h")
  run_name = f"{FLAGS.game_name}__{FLAGS.exp_name}__"
  if FLAGS.game_name == "atari":
    run_name += f"{FLAGS.gym_id}__"
  run_name += f"{FLAGS.seed}__{current_month_text}__{current_day}__{int(time.time())}"

  writer = SummaryWriter(f"runs/{run_name}")
  writer.add_text(
      "hyperparameters",
      "|param|value|\n|-|-|\n%s" %
      ("\n".join([f"|{key}|{value}|" for key, value in vars(FLAGS).items()])),
  )

  random.seed(FLAGS.seed)
  np.random.seed(FLAGS.seed)
  torch.manual_seed(FLAGS.seed)
  torch.backends.cudnn.deterministic = FLAGS.torch_deterministic

  device = torch.device(
      "cuda" if torch.cuda.is_available() and FLAGS.cuda else "cpu")
  logging.info("Using device: %s", str(device))

  if FLAGS.game_name == "atari":
    envs = SyncVectorEnv([
        make_single_atari_env(FLAGS.gym_id, FLAGS.seed + i, i, False,
                              run_name)() for i in range(FLAGS.num_envs)
    ])
    agent_fn = PPOAtariAgent
  else:
    envs = SyncVectorEnv([
        make_single_env(FLAGS.game_name, FLAGS.seed + i)()
        for i in range(FLAGS.num_envs)
    ])
    agent_fn = PPOAgent

  game = envs.envs[0]._game  # pylint: disable=protected-access
  info_state_shape = game.observation_tensor_shape()

  num_updates = FLAGS.total_timesteps // batch_size
  agent = PPO(
      input_shape=info_state_shape,
      num_actions=game.num_distinct_actions(),
      num_players=game.num_players(),
      player_id=0,
      num_envs=FLAGS.num_envs,
      steps_per_batch=FLAGS.num_steps,
      num_minibatches=FLAGS.num_minibatches,
      update_epochs=FLAGS.update_epochs,
      learning_rate=FLAGS.learning_rate,
      gae=FLAGS.gae,
      gamma=FLAGS.gamma,
      gae_lambda=FLAGS.gae_lambda,
      normalize_advantages=FLAGS.norm_adv,
      clip_coef=FLAGS.clip_coef,
      clip_vloss=FLAGS.clip_vloss,
      entropy_coef=FLAGS.ent_coef,
      value_coef=FLAGS.vf_coef,
      max_grad_norm=FLAGS.max_grad_norm,
      target_kl=FLAGS.target_kl,
      device=device,
      writer=writer,
      agent_fn=agent_fn,
  )

  n_reward_window = 50
  recent_rewards = collections.deque(maxlen=n_reward_window)
  time_step = envs.reset()
  for update in range(num_updates):
    for _ in range(FLAGS.num_steps):
      agent_output = agent.step(time_step)
      time_step, reward, done, unreset_time_steps = envs.step(
          agent_output, reset_if_done=True)

      if FLAGS.game_name == "atari":
        # Get around the fact that
        # stable_baselines3.common.atari_wrappers.EpisodicLifeEnv will modify
        # rewards at the LIFE and not GAME level by only counting
        # rewards of finished episodes
        for ts in unreset_time_steps:
          info = ts.observations.get("info")
          if info and "episode" in info:
            real_reward = info["episode"]["r"]
            writer.add_scalar("charts/player_0_training_returns", real_reward,
                              agent.total_steps_done)
            recent_rewards.append(real_reward)
      else:
        for ts in unreset_time_steps:
          if ts.last():
            real_reward = ts.rewards[0]
            writer.add_scalar("charts/player_0_training_returns", real_reward,
                              agent.total_steps_done)
            recent_rewards.append(real_reward)

      agent.post_step(reward, done)

    if FLAGS.anneal_lr:
      agent.anneal_learning_rate(update, num_updates)

    agent.learn(time_step)

    if update % FLAGS.eval_every == 0:
      logging.info("-" * 80)
      logging.info("Step %s", agent.total_steps_done)
      logging.info("Summary of past %i rewards\n %s",
                   n_reward_window,
                   pd.Series(recent_rewards).describe())

  writer.close()
  logging.info("All done. Have a pleasant day :)")


if __name__ == "__main__":
  app.run(main)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for pyspiel Chat Game."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python.games import chat_game  # pylint: disable=unused-import

from open_spiel.python.games.chat_games.configs import config_fixed_mock
from open_spiel.python.games.chat_games.configs import config_rnd_mock

from open_spiel.python.games.chat_games.utils import test_utils as chat_test_utils

import pyspiel


GLOBAL_TEST_LLM = chat_test_utils.TestLLM.MOCK


class ChatGameTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.fixed_config = config_fixed_mock.get_config()
    self.random_config = config_rnd_mock.get_config()

    vectorizer = chat_test_utils.MockVectorizer()
    self.vectorize = vectorizer.vectorize

  @parameterized.named_parameters(
      dict(testcase_name='fixed_scenario', fixed_scenario=True),
      dict(testcase_name='random_scenario', fixed_scenario=False))
  def test_game_from_cc(self, fixed_scenario):
    """Runs our standard game tests, checking API consistency."""

    if fixed_scenario:
      config = self.fixed_config
    else:
      config = self.random_config

    game = pyspiel.load_game('chat_game', config.params.to_dict())

    game.load_chat_game(llm_type=GLOBAL_TEST_LLM,
                        vectorize=self.vectorize,
                        seed=1234,
                        **config.game)

    pyspiel.random_sim_test(game, num_sims=10, serialize=False, verbose=True)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A pyspiel config for meta-generated meeting schedule negotiation games.
"""

import collections

from ml_collections import config_dict

from open_spiel.python.games.chat_games.envs.base_envs import schedule_meeting_with_tone_info as env_schedule_meeting_with_tone_info
from open_spiel.python.games.chat_games.envs.observations import summary
from open_spiel.python.games.chat_games.envs.observations import utils as obs_utils
from open_spiel.python.games.chat_games.envs.payoffs import schedule_meeting as payoffs_schedule_meeting
from open_spiel.python.games.chat_games.envs.scenarios.domains import schedule_meeting as scenario_schedule_meeting


def get_config():
  """Get configuration for chat game."""
  config = config_dict.ConfigDict()

  num_players = 2

  observations = [
      obs_utils.Observation(summary.PREFIX, summary.POSTFIX)
      for _ in range(num_players)
  ]

  header = env_schedule_meeting_with_tone_info.HEADER

  payoffs = [payoffs_schedule_meeting.PAYOFF]

  given_prompt_actions = collections.OrderedDict()
  tones = ['calm',
           'assertive',
           'submissive',
           'any']
  given_prompt_actions[header.action_keys[0]] = tones
  num_tones = len(tones)

  given_private_info = collections.OrderedDict()
  given_private_info['day_prefs'] = [scenario_schedule_meeting.DAY_PREFS_A,
                                     scenario_schedule_meeting.DAY_PREFS_B]
  given_private_info['ooo_days'] = [scenario_schedule_meeting.OOO_A,
                                    scenario_schedule_meeting.OOO_B]

  scenario_a = env_schedule_meeting_with_tone_info.Scenario(
      scenario_schedule_meeting.SCENARIO_A,
      'Bob',
      'Suzy',
      scenario_schedule_meeting.OOO_A,
      scenario_schedule_meeting.DAY_PREFS_A,
      'calm')

  llm_termination_prompt = scenario_schedule_meeting.LLM_TERMINATION_PROMPT

  params = {'num_distinct_actions': num_players * num_tones,
            'num_llm_seeds': 2,
            'num_init_states': 1,
            'num_players': num_players,
            'min_utility': min([float(p.min) for p in payoffs]),
            'max_utility': max([float(p.max) for p in payoffs]),
            'num_max_replies': 1,
            'silence_logging': True}

  config.params = params

  config.game = config_dict.ConfigDict()
  config.game.observations = observations
  config.game.header = header
  config.game.payoffs = payoffs
  config.game.given_prompt_actions = given_prompt_actions
  config.game.num_private_info = (2, 2)
  config.game.given_names = ['Bob', 'Suzy']
  config.game.given_private_info = given_private_info
  config.game.initial_scenario = scenario_a
  config.game.llm_list_suffix = 'Output: '
  config.game.llm_termination_prompt = llm_termination_prompt

  return config



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for computing gradient information: run games and record payoffs.
"""

from absl import logging  # pylint:disable=unused-import

import numpy as np


def construct_game_queries(base_profile, num_checkpts):
  """Constructs a list of checkpoint selection tuples to query value function.

  Each query tuple (p1's selected checkpt, ..., p7's selected checkpt)
  fixes the players in the game of diplomacy to be played. It may be necessary
  to play several games with the same players to form an accurate estimate of
  the value or payoff for each player as checkpts contain stochastic policies.

  Args:
    base_profile: list of selected checkpts for each player, i.e.,
      a sample from the player strategy profile ([x_i ~ p(x_i)])
    num_checkpts: number of checkpts available to each player
  Returns:
    Set of query tuples containing a selected checkpoint index for each player.
  """
  new_queries = set([])

  pi, pj = 0, 1
  new_profile = list(base_profile)
  for ai in range(num_checkpts):
    new_profile[pi] = ai
    for aj in range(num_checkpts):
      new_profile[pj] = aj
      query = tuple(new_profile)
      new_queries.update([query])

  return new_queries


def construct_game_queries_for_exp(base_profile, num_checkpts):
  """Constructs a list of checkpoint selection tuples to query value function.

  Each query tuple (p1's selected checkpt, ..., p7's selected checkpt)
  fixes the players in the game of diplomacy to be played. It may be necessary
  to play several games with the same players to form an accurate estimate of
  the value or payoff for each player as checkpts contain stochastic policies.

  Args:
    base_profile: list of selected checkpts for each player, i.e.,
      a sample from the player strategy profile ([x_i ~ p(x_i)])
    num_checkpts: number of checkpts available to each player
  Returns:
    Set of query tuples containing a selected checkpoint index for each player.
  """
  new_queries = set([])

  pi = 0
  new_profile = list(base_profile)
  for ai in range(num_checkpts):
    new_profile[pi] = ai
    query = tuple(new_profile)
    new_queries.update([query])

  return new_queries


def run_games_and_record_payoffs(game_queries, evaluate_game, ckpt_to_policy):
  """Simulate games according to game queries and return results.

  Args:
    game_queries: set of tuples containing indices specifying each players strat
    evaluate_game: callable function that takes a list of policies as argument
    ckpt_to_policy: maps a strat (or checkpoint) to a policy
  Returns:
    dictionary: key=query, value=np.array of payoffs (1 for each player)
  """
  game_results = {}
  for query in game_queries:
    policies = [ckpt_to_policy[ckpt] for ckpt in query]
    payoffs = evaluate_game(policies)
    game_results.update({query: payoffs})
  return game_results


def form_payoff_matrices(game_results, num_checkpts):
  """Packages dictionary of game results into a payoff tensor.

  Args:
    game_results: dictionary of payoffs for each game evaluated
    num_checkpts: int, number of strats (or ckpts) per player
  Returns:
    payoff_matrices: np.array (2 x num_checkpts x num_checkpts) with payoffs for
      two players (assumes symmetric game and only info for 2 players is needed
      for stochastic gradients)
  """
  payoff_matrices = np.zeros((2, num_checkpts, num_checkpts))
  for profile, payoffs in game_results.items():
    i, j = profile[:2]
    payoff_matrices[:, i, j] = payoffs[:2]
  return payoff_matrices



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.voting.plurality."""

from absl.testing import absltest

from open_spiel.python.voting import base
from open_spiel.python.voting import copeland


class CopelandVotingTest(absltest.TestCase):
  def test_copeland_construction(self):
    method = copeland.CopelandVoting()
    self.assertEqual(method.name(), "copeland")

  def test_copeland_basic_run(self):
    votes = [["a", "b", "c"], ["a", "c", "b"], ["b", "a", "c"]]
    profile = base.PreferenceProfile(votes=votes)
    method = copeland.CopelandVoting()
    outcome = method.run_election(profile)
    self.assertListEqual(outcome.ranking, ["a", "b", "c"])
    self.assertListEqual(outcome.scores, [2.0, 1.0, 0.0])

  def test_copeland_basic_run2(self):
    votes = [
        base.WeightedVote(1, ["a", "b", "c"]),
        base.WeightedVote(2, ["a", "c", "b"]),
        base.WeightedVote(3, ["b", "a", "c"]),
    ]
    profile = base.PreferenceProfile(votes=votes)
    method = copeland.CopelandVoting()
    outcome = method.run_election(profile)
    self.assertTrue(outcome.ranking == ["a", "b", "c"] or
                    outcome.ranking == ["b", "a", "c"])
    self.assertListEqual(outcome.scores, [1.5, 1.5, 0.0])


if __name__ == "__main__":
  absltest.main()
